/* This file is part of mailfrom filter.
   Copyright (C) 2005, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <resolv.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <pthread.h>
#include <mailutils/mailutils.h>
#include "mailfrom.h"

struct mx_buffer {
	unsigned pref;
	char *name;
};

static int
comp_pref(const void *a, const void *b)
{
	const struct mx_buffer *ma = a, *mb = b;
	if (ma->pref > mb->pref)
		return 1;
	else if (ma->pref < mb->pref)
		return -1;
	return 0;
}

static void
free_mx_buffer(struct mx_buffer *mx_buffer, int nmx)
{
	int i;
	for (i = 0; i < nmx; i++)
		free(mx_buffer[i].name);
}

static getmx_status
_getmx(char *host, char *answer, size_t answer_size, mxbuf_t mxbuf)
{
	int n, nmx, i, count;
	ns_msg msg;
	struct mx_buffer mx_buffer[MAXMXCOUNT];

	debug(100,("Getting MX records for %s", host));
	n = res_query (host, C_IN, T_MX, answer, answer_size);
	if (n < 0) {
		debug(10, ("res_query failed (errno=%s, h_errno=%d)",
			   mu_strerror(errno), h_errno));
		switch (h_errno) {
		case NO_DATA:
		case NO_RECOVERY:
		case HOST_NOT_FOUND:
			return getmx_not_found;

		case TRY_AGAIN:
		case -1:
			return getmx_temp_failure;

		default:
			mu_error("res_nsearch(%s) failed with unexpected h_errno %d",
				 host, h_errno);
			return getmx_failure;
		}
	}

	if (ns_initparse(answer, n, &msg) < 0) {
		debug(2,("ns_initparse(host %s) failed", host));
		return getmx_failure;
	}
		
	/* Get number of MXs */
	count = ns_msg_count(msg, ns_s_an);
	if (count == 0) 
		return getmx_not_found;

	/* Collect MX records */
	for (i = nmx = 0; i < count; i++) {
		ns_rr rr;
		
		if (ns_parserr(&msg, ns_s_an, i, &rr) < 0) {
			debug(2,("ns_parserrd(%d) failed (host %s)",
				  i, host));
			free_mx_buffer(mx_buffer, nmx);
			return getmx_failure;
		}
		if (ns_rr_type(rr) == ns_t_mx) {
			int s;
			char tname[NS_MAXDNAME];
			const u_char *rdata;
			int rdlen;

			rdata = ns_rr_rdata(rr);
			rdlen = ns_rr_rdlen(rr);

			s = dn_expand(ns_msg_base(msg), ns_msg_end(msg),
				      rdata+NS_INT16SZ, tname, sizeof tname);
			if (s < 0) {
				debug(2,("dn_expand failed"));
				free_mx_buffer(mx_buffer, nmx);
				return getmx_failure;
			}
			mx_buffer[nmx].pref = ns_get16(rdata);
			mx_buffer[nmx].name = strdup(tname);
			debug(100,("MX %u %s", mx_buffer[nmx].pref,
				   mx_buffer[nmx].name));
			if (nmx++ > MAXMXCOUNT)
				break;
		}
	}

	/* Sort according to preference value */
	qsort(mx_buffer, nmx, sizeof mx_buffer[0], comp_pref);

	/* Prepare return value */
	memset(mxbuf, 0, sizeof(mxbuf_t));
	for (i = 0; i < nmx; i++) 
		mxbuf[i] = mx_buffer[i].name;
	return getmx_success;
}

static int
is_ipaddr(char *str)
{
	for (; *str; str++)
		if (!(isdigit(*str) || *str == '.'))
			return 0;
	return 1;
}

static pthread_mutex_t getmx_mutex = PTHREAD_MUTEX_INITIALIZER;

getmx_status
getmx(char *ipstr, mxbuf_t mxbuf)
{
	struct sockaddr_in sa;
	char hbuf[NI_MAXHOST];
	getmx_status status = getmx_failure;

	pthread_mutex_lock(&getmx_mutex);
	memset(&sa, 0, sizeof sa);
	sa.sin_family = AF_INET;
	if (inet_aton(ipstr, &sa.sin_addr)
	    && getnameinfo((struct sockaddr *)&sa, sizeof sa,
			   hbuf, sizeof(hbuf),
			   NULL, 0,
			   0) == 0
	    && !is_ipaddr(hbuf)) {
		unsigned char *answer = malloc(MAXPACKET);
		if (!answer)
			status = getmx_failure;
		else {
			char *p;

			for (p = strchr(hbuf, '.'); p; p = strchr(p+1, '.')) {
				status = _getmx(p+1, answer, MAXPACKET, mxbuf);
				if (status == getmx_success)
					break;
			}
			free(answer);
		}
	}
	pthread_mutex_unlock(&getmx_mutex);
	return status;
}
		
