/* This file is part of mailfrom filter.
   Copyright (C) 2005, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include <syslog.h>
#include <pwd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>

#ifdef HAVE_SYSEXITS_H
# include <sysexits.h>
#else
# define EX_OK          0       /* successful termination */
# define EX__BASE       64      /* base value for error messages */
# define EX_USAGE       64      /* command line usage error */
# define EX_DATAERR     65      /* data format error */
# define EX_NOINPUT     66      /* cannot open input */
# define EX_NOUSER      67      /* addressee unknown */
# define EX_NOHOST      68      /* host name unknown */
# define EX_UNAVAILABLE 69      /* service unavailable */
# define EX_SOFTWARE    70      /* internal software error */
# define EX_OSERR       71      /* system error (e.g., can't fork) */
# define EX_OSFILE      72      /* critical OS file missing */
# define EX_CANTCREAT   73      /* can't create (user) output file */
# define EX_IOERR       74      /* input/output error */
# define EX_TEMPFAIL    75      /* temp failure; user is invited to retry */
# define EX_PROTOCOL    76      /* remote error in protocol */
# define EX_NOPERM      77      /* permission denied */
# define EX_CONFIG      78      /* configuration error */
# define EX__MAX        78      /* maximum listed value */
#endif

#include <mailutils/mailutils.h>
#include <mailutils/argp.h>
#include <libmilter/mfapi.h>

#include "mailfrom.h"

/* Configurable return values */
sfsistat status_failure = SMFIS_CONTINUE;
/* Internal failure.
   Allowed values: SMFIS_ACCEPT, SMFIS_CONTINUE, SMFIS_DISCARD, SMFIS_REJECT */

sfsistat status_bad_sender = SMFIS_DISCARD;
/* Sender validity is not confirmed.
   Allowed values: SMFIS_DISCARD, SMFIS_REJECT */

sfsistat status_temp_failure = SMFIS_TEMPFAIL;
/* Temporary failure (4xx code).
   Allowed values: SMFIS_ACCEPT, SMFIS_CONTINUE, SMFIS_DISCARD, SMFIS_REJECT,
                   SMFIS_TEMPFAIL */

int do_transcript;  /* Enable session transript */
int debug_level;    /* Debugging level */
int test_mode;      /* Enable test mode */
int log_to_stderr;  /* Use stderr for logging */
char *smtp_domain;  /* Default SMTP domain for EHLO command */
char *portspec = DEFAULT_SOCKET; /* Communication socket specification */
int force_remove;   /* Remove local communication socket if it already
		       exists */
int foreground;     /* Stay in foreground */

char *pidfile = DEFAULT_PIDFILE;
char *user = DEFAULT_USER;       /* Switch to this user privileges after startup */   
/* I/O settings */
unsigned io_timeout = 3;
unsigned io_attempts = 3;

/* FIXME: These should be configurable, too */
char *reject_code = "550";
char *reject_xcode = NULL;
char *reject_message = "Sender validity not confirmed by MX servers";


/* Logging & debugging */
static const char *sfsistat_str(sfsistat stat);

int
syslog_printer (int prio, const char *fmt, va_list ap)
{
#ifdef HAVE_VSYSLOG
  vsyslog (prio, fmt, ap);
#else
  char buf[128];
  vsnprintf (buf, sizeof buf, fmt, ap);
  syslog (prio, "%s", buf);
#endif
  return 0;
}

int
syslog_error_printer (const char *fmt, va_list ap)
{
	return syslog_printer(LOG_ERR, fmt, ap);
}

void
vlogmsg(int prio, char *fmt, va_list ap)
{
	if (log_to_stderr) {
		vfprintf(stderr, fmt, ap);
		fprintf(stderr, "\n");
	} else
		syslog_printer(prio, fmt, ap);
}

void
logmsg(int prio, char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(prio, fmt, ap);
	va_end(ap);
}

void
transcript(char *prefix, char *msg)
{
	if (do_transcript) {
		int len = strlen(msg);
		if (msg[len-1] == '\n') {
			--len;
			if (len > 0 && msg[len-1] == '\r')
				--len;
		}
		if (len)
			logmsg(LOG_INFO, "%s %*.*s", prefix, len, len, msg);
	}
}

void
debug_log(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(LOG_DEBUG, fmt, ap);
	va_end(ap);
}


/* Sendmail class file support */

static list_t domain_list;

static int
compare_string(const void *item, const void *value)
{
	return strcmp(item, value);
}

void
read_domain_file(char *name)
{
	FILE *fp;
	char buf[256];
	char *p;
	
	fp = fopen(name, "r");
	if (!fp) {
		mu_error("cannot open file `%s': %s",
			 name, mu_strerror(errno));
		return;
	}

	if (!domain_list) {
		list_create(&domain_list);
		list_set_comparator(domain_list, compare_string);
	}

	while (p = fgets(buf, sizeof buf, fp)) {
		char *q;
		
		while (*p && isspace(*p))
			p++;
		if (*p == '#')
			continue;

		for (q = p + strlen(p) - 1; q > p && isspace(*q); q--)
			*q = 0;

		if (*p == 0)
			continue;
		
		list_append(domain_list, strdup(p));
	}

	fclose(fp);
}

int
relayed_domain_p(char *name)
{
	char *p;

	for (p = strchr(name, '.'); p; p = strchr(p+1, '.'))
		if (list_locate(domain_list, p+1, NULL) == 0) {
			debug(10,("%s is in relayed domain %s",
				  name, p+1));
			return 1;
		}
	return 0;
}

int
host_in_relayed_domain_p(char *client)
{
	int rc;
	struct sockaddr_in sa;
	char hbuf[NI_MAXHOST];
	
	if (list_is_empty(domain_list))
		return 0;

	memset(&sa, 0, sizeof sa);
	sa.sin_family = AF_INET;
	if (inet_aton(client, &sa.sin_addr) == 0)
		return 0;
	rc = getnameinfo((struct sockaddr *)&sa, sizeof sa,
			 hbuf, sizeof(hbuf),
			 NULL, 0,
			 0);
	if (rc)
		return 0;
	return relayed_domain_p(hbuf);
}


/* SMTP I/O functions */

#define MLFIPRIV(ctx)  ((struct smtp_io_data*) smfi_getpriv(ctx))

struct smtp_io_data {
	SMFICTX *ctx;        /* Milter context */
	stream_t stream;     /* I/O stream */
	size_t send_off;     /* Send offset */
	size_t recv_off;     /* Receive offset */
	struct obstack stk;  /* Obstack for keeping commands/replies */
	char *command;       /* Last issued command */ 
	char *reply;         /* Last received reply */
	char *start;         /* First line of the reply, if it was multiline */
	char buf[128];       /* Input buffer */
	size_t level;        /* Number of bytes in buf */
	int code;            /* Reply code */
};

void
smtp_io_data_init(struct smtp_io_data *dat, SMFICTX *ctx, stream_t stream)
{
	obstack_init(&dat->stk);
	dat->send_off = dat->recv_off = 0;
	dat->stream = stream;
	dat->start = dat->command = dat->reply = NULL;
	dat->level = 0;
	dat->ctx = ctx;
	if (ctx) 	/* save the private data */
		smfi_setpriv(ctx, dat);
}

void
smtp_io_data_destroy(struct smtp_io_data *dat)
{
	if (dat) {
		stream_close (dat->stream);
		stream_destroy(&dat->stream, stream_get_owner(dat->stream));
		obstack_free(&dat->stk, NULL);
		if (dat->ctx)
			smfi_setpriv(dat->ctx, NULL);
	}
}

void
smtp_stream_wait(stream_t stream, int flags)
{
	struct timeval tv;
	tv.tv_sec = io_timeout;
	tv.tv_usec = 0;
	stream_wait(stream, &flags, &tv);
}

void
smtp_wait(struct smtp_io_data *dat, int flags)
{
	return smtp_stream_wait(dat->stream, flags);
}

int
smtp_send(struct smtp_io_data *dat, char *command)
{
	int attempt = 0;
	size_t len = strlen(command);

	transcript("SEND:", command);
	do {
		size_t nb;
		int rc = stream_write(dat->stream, command, len,
				      dat->send_off, &nb);
		if (rc == 0) {
			dat->send_off += nb;
			len -= nb;
			command += nb;
		} else if (rc == EAGAIN) {
			if (attempt > io_attempts) { 
				mu_error ("stream_write timed out");
				return -1;
			} else {
				smtp_wait(dat, MU_STREAM_READY_WR);
				attempt++;
				continue;
			}
		} else {
			mu_error ("stream_write: %s", mu_strerror (rc));
			return -1;
		}
	} while (len > 0);
	return 0;
}

int
smtp_send2(struct smtp_io_data *dat, char *command, char *arg)
{
	int rc;
	
	obstack_grow(&dat->stk, command, strlen(command));
	if (arg)
		obstack_grow(&dat->stk, arg, strlen(arg));
	obstack_1grow(&dat->stk, 0);
	dat->command = obstack_finish(&dat->stk);
	
	rc = smtp_send(dat, dat->command);
	if (rc == 0)
		rc = smtp_send(dat, "\r\n");
	return rc;
}

int
smtp_send3(struct smtp_io_data *dat, char *command, char *arg1, char *arg2)
{
	int rc;
	
	obstack_grow(&dat->stk, command, strlen(command));
	obstack_grow(&dat->stk, arg1, strlen(arg1));
	obstack_grow(&dat->stk, arg2, strlen(arg2));
	obstack_1grow(&dat->stk, 0);
	dat->command = obstack_finish(&dat->stk);
	
	rc = smtp_send(dat, dat->command);
	if (rc == 0)
		rc = smtp_send(dat, "\r\n");
	
	return rc;
}

int
smtp_recvline(struct smtp_io_data *dat)
{
	size_t attempt = 0;
	for (;;) {
		char *p;
		
		if (dat->level == 0) {
			int rc = stream_read(dat->stream,
					     dat->buf, sizeof dat->buf,
					     dat->recv_off, &dat->level);
			if (rc == 0) {
				if (dat->level == 0) {
					mu_error("stream_read: read 0 bytes");
					return -1;
				}
				dat->recv_off += dat->level;
			} else if (rc == EAGAIN) {
				if (attempt > io_attempts) {
					mu_error("stream_read timed out");
					return -1;
				}
				smtp_wait(dat, MU_STREAM_READY_RD);
				continue;
			} else {
				mu_error ("stream_read: %s", mu_strerror (rc));
				return -1;
			}
		}
		attempt = 0;
		
		p = memchr(dat->buf, '\n', dat->level);
		if (!p) {
			obstack_grow(&dat->stk, dat->buf, dat->level);
			dat->level = 0;
			continue;
		} else {
			size_t len = p - dat->buf + 1;
			obstack_grow(&dat->stk, dat->buf, len);
			obstack_1grow(&dat->stk, 0);
			dat->reply = obstack_finish(&dat->stk);
			dat->level -= len;
			memmove(dat->buf, dat->buf + len, dat->level);
			break;
		}
	}
	return 0;
}

int
smtp_recv(struct smtp_io_data *dat)
{
	char *p;
	dat->start = NULL;
	do {
		int code;
		int rc = smtp_recvline(dat);
		if (rc)
			return -1;
		transcript("RECV:", dat->reply);
		code = strtoul (dat->reply, &p, 0);
		if (p - dat->reply != 3 || (*p != '-' && *p != ' ')) {
			mu_error("Unexpected reply from server: %s",
				 dat->reply);
			return -1;
		} else if (!dat->start) {
			dat->start = dat->reply;
			dat->code = code;
		} else if (dat->code != code) {
			mu_error("Unexpected reply code from server: %d",
				 code);
			return -1;
		}
	} while (*p == '-');
	obstack_1grow (&dat->stk, 0);
	obstack_finish (&dat->stk);
	return 0;
}


/* Milter-specific functions */

void
log_status(sfsistat status, SMFICTX *ctx, struct smtp_io_data *io)
{
	if (debug_level >= 1) {
		char *id = smfi_getsymval(ctx, "i");
		const char *str = sfsistat_str(status);
		if (status != SMFIS_ACCEPT && io) {
			if (str)
				logmsg(LOG_INFO,
				       "%s: %s, command=%s, reply=%s",
				       id ? id : "(nil)", str,
				       io->command, io->reply);
			else
				logmsg(LOG_INFO,
				       "%s: status %d, command=%s, reply=%s",
				       id ? id : "(nil)", status,
				       io->command, io->reply);
		} else {
			if (str)
				logmsg(LOG_INFO,
				       "%s: %s", id ? id : "(nil)", str);
			else
				logmsg(LOG_INFO,
				       "%s: status %d",
				       id ? id : "(nil)", status);
		}
	}
}

getmx_status
check_on_host(SMFICTX *ctx, char *email, char *client_addr)
{
	int rc;
	stream_t stream;
	struct smtp_io_data io;
	getmx_status status = getmx_success;
	int count = 0;
	
	debug(10, ("email = %s, client_addr = %s", email, client_addr));
	rc = tcp_stream_create (&stream, client_addr, 25, MU_STREAM_NONBLOCK);
	if (rc) {
		mu_error ("tcp_stream_create: %s", mu_strerror (rc));
		return getmx_temp_failure;
	}

	while (rc = stream_open(stream)) {
		if ((rc == EAGAIN || rc == EINPROGRESS)
		    && count <= io_attempts) {
			count++;
			smtp_stream_wait(stream, MU_STREAM_READY_WR);
			continue;
		}
		mu_error("stream_open: %s, %d", mu_strerror(rc), count);
		stream_destroy(&stream, stream_get_owner(stream));
		if (rc == EAGAIN || rc == EINPROGRESS)
			return getmx_temp_failure;
		else
			return getmx_failure;
	}
	debug(100,("stream opened"));
	
	smtp_io_data_init(&io, ctx, stream);

	do {
		char *buf;
		char *descr;
		if (smtp_recv(&io) || io.code != 220) {
			status = getmx_not_found;
			break;
		}
		smtp_send2(&io, "HELO ", smtp_domain);
		if (smtp_recv(&io) || io.code != 250) {
			status = getmx_not_found;
			break;
		}
		smtp_send2(&io, "MAIL FROM: <>", NULL);
		if (smtp_recv(&io) || io.code != 250) {
			status = getmx_not_found;
			break;
		}
		smtp_send3(&io, "RCPT TO: <", email, ">");
		if (smtp_recv(&io) || (io.code != 250 && io.code / 100 != 4)) {
			status = getmx_not_found;
			break;
		}
		smtp_send2(&io, "QUIT", NULL);
	} while (0);

	smtp_io_data_destroy(&io);

	return status;
}

sfsistat
check_sender(SMFICTX *ctx, char *email, char *client_addr)
{
	getmx_status rc = status_bad_sender;
	sfsistat status;
	
	if ((rc = check_on_host(ctx, email, client_addr)) != getmx_success) {
		int i;
		mxbuf_t mxbuf;
		getmx_status mxstat = getmx(client_addr, mxbuf);

		switch (mxstat) {
		case getmx_temp_failure:
			rc = status_temp_failure;
			break;

		case getmx_failure:
		case getmx_not_found:
			rc = getmx_not_found;
			break;

		case getmx_success:
			debug(2,("Checking MX servers for %s", email));
			for (i = 0; i < MAXMXCOUNT && mxbuf[i]; i++) {
				rc = check_on_host(ctx, email, mxbuf[i]);
				if (rc == getmx_success
				    || rc == getmx_not_found)
					break;
				free(mxbuf[i]);
			}
			for (; i < MAXMXCOUNT && mxbuf[i]; i++) 
				free(mxbuf[i]);
		}
		
	}

	switch (rc) {
	case getmx_temp_failure:
	case getmx_failure:
		status = status_temp_failure;
		break;

	case getmx_not_found:
		status = status_bad_sender;
		break;

	case getmx_success:
		status = SMFIS_ACCEPT;
	}
	log_status(status, ctx, NULL);
	return status;
}

sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	char *mail_addr = smfi_getsymval(ctx, "f");
	char *client = smfi_getsymval(ctx, "{client_addr}");
	sfsistat status;
	
	if (!mail_addr) {
		mu_error("$f is not available. Fix your confMILTER_MACROS_ENVFROM settings");
		return status_failure;
	}
	if (mail_addr[0] == 0)
		return SMFIS_ACCEPT;
	if (!client) {
		mu_error("${client_addr} is not available. Fix your confMILTER_MACROS_ENVFROM settings");
		return status_failure;
	}

	if (host_in_relayed_domain_p(client)) 
		return SMFIS_ACCEPT;
	
	status = check_sender(ctx, mail_addr, client);

	if (status == SMFIS_REJECT) 
		smfi_setreply(ctx, reject_code, reject_xcode, reject_message);
	
	return status;
}

void
filter_cleanup(struct smtp_io_data *dat)
{
	debug(100,("cleaning up"));
	if (!foreground)
		daemon_remove_pidfile();
	smtp_io_data_destroy(dat);
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{
	filter_cleanup(MLFIPRIV(ctx));
}		
		

struct smfiDesc smfilter =
{
	"MailfromFilter",	
	SMFI_VERSION,	
	0,
	NULL,	/* connection info filter */
	NULL,	/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	NULL,	/* envelope recipient filter */
	NULL,	/* header filter */
	NULL,	/* end of header */
	NULL,	/* body block filter */
	NULL,	/* end of message */
	mlfi_abort,	/* message aborted */
	NULL,	/* connection cleanup */
};


/* Command line parsing */
const char *program_version = "mailfrom (" PACKAGE_STRING ")";
static char doc[] = N_("mailfrom -- MAIL FROM milter checker");
static char args_doc[] = "";

#define OPTION_MILTER_TIMEOUT 256
#define OPTION_SHOW_DEFAULTS 257
#define OPTION_FOREGROUND 258
#define OPTION_PIDFILE 259
#define OPTION_DOMAIN_FILE 260
#define OPTION_TIMEOUT 261
#define OPTION_RETRY 262

static struct argp_option options[] = {
#define GRP 10
	{ NULL, 0, NULL, 0,
	  "General options", GRP },
	{"domain", 'D', "STRING", 0,
	 "Set default SMTP domain", GRP+1},
	{"port", 'p', "STRING", 0,
	 "Set communication socket", GRP+1},
	{"remove", 'r', NULL, 0,
	 "Forse removing of the local socket file, if it already exists",
	 GRP+1},
	{"foreground", OPTION_FOREGROUND, NULL, 0,
	 "Stay in foreground", GRP+1},
	{"pidfile", OPTION_PIDFILE, "FILE", 0,
	 "Set pidfile name", GRP+1},
	{"user", 'u', "NAME", 0,
	 "Switch to this user privileges after startup", GRP+1},

#undef GRP
#define GRP 20
	{ NULL, 0, NULL, 0,
	  "Action control", GRP },
	{"action-failure", 'F', "ACTION",  0,
	 "Action to take on failure (continue,discard,reject). Default is continue.", GRP+1},
	{"action-bad-sender", 'B', "ACTION",  0,
	 "Action to take of sender validity is not confirmed (discard,reject). Default is discard.", GRP+1},
	{"action-temp-failure", 'T', "ACTION", 0,
	 "Action to take on temporary failures (tempfail,accept,discard,reject). Default is tempfail",
	 GRP+1},

#undef GRP
#define GRP 30
	{ NULL, 0, NULL, 0,
	  "Timeout control", GRP },
	{"milter-timeout", OPTION_MILTER_TIMEOUT, 0, 0,
	 "Set MTA connection timeout", GRP+1 },
	{"timeout", OPTION_TIMEOUT, "NUMBER", 0,
	 "Set I/O operation timeout (seconds)", GRP+1},
	{"retry", OPTION_RETRY, "NUMBER", 0,
	 "Retry failed I/O operations NUMBER times", GRP+1},

	{"relayed-domain-file", OPTION_DOMAIN_FILE, "FILE", 0,
	 "Read relayed domains from FILE", GRP+1},
	
#undef GRP
#define GRP 40
	{ NULL, 0, NULL, 0,
	  "Informational and debugging options", GRP },
	{"transcript", 'X', NULL, 0,
	 "Enable transcript", GRP+1},
	{"debug", 'd', "LEVEL", 0,
	 "Set debugging level", GRP+1},
	{"stderr", 's', NULL, 0,
	 "Log to stderr", GRP+1},
	{"test", 't', NULL, 0,
	 "Run in test mode", GRP+1},
	{"show-defaults", OPTION_SHOW_DEFAULTS, NULL, 0,
	 "Show compilation defaults", GRP+1},
#undef GRP
	NULL
};

#define MAP(c) (1<<(c))

struct sfsistat_tab {
	char *name;
	sfsistat stat;
} sfsistat_tab[] = {
	{ "accept", SMFIS_ACCEPT },
	{ "continue", SMFIS_CONTINUE },
	{ "discard", SMFIS_DISCARD },
	{ "reject", SMFIS_REJECT },
	{ "tempfail", SMFIS_TEMPFAIL },
	NULL
};

static const char *
sfsistat_str(sfsistat stat)
{
	struct sfsistat_tab *p;
	for (p = sfsistat_tab; p->name; p++)
		if (p->stat == stat) 
			return p->name;
	return NULL;
}

void
print_stat(sfsistat stat)
{
	struct sfsistat_tab *p;
	for (p = sfsistat_tab; p->name; p++)
		if (p->stat == stat) {
			printf("%s", p->name);
			return;
		}
	printf("%d", stat);
}

#define STAT_PREFIX "SMFIS_"
int
parse_action(char *code, sfsistat *act, unsigned long map)
{
	struct sfsistat_tab *p;
	
	if (strncasecmp(code, STAT_PREFIX, sizeof STAT_PREFIX - 1) == 0)
		code += sizeof STAT_PREFIX - 1;
	for (p = sfsistat_tab; p->name; p++)
		if (strcasecmp(p->name, code) == 0) {
			if ((MAP(p->stat) & map) == 0)
				return 1;
			*act = p->stat;
			return 0;
		}
	return 1;
}

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
	switch (key) {
	case 'F':
		if (parse_action(arg, &status_failure,
				 MAP(SMFIS_ACCEPT)|MAP(SMFIS_CONTINUE)|
				 MAP(SMFIS_DISCARD)|MAP(SMFIS_REJECT)))
			argp_error(state,
				   "Bad value for '--failure-action' option");
		break;
		
	case 'B':
		if (parse_action(arg, &status_bad_sender,
				 MAP(SMFIS_DISCARD)|MAP(SMFIS_REJECT)))
			argp_error(state,
				   "Bad value for '--bad-sender-action-action' option");
		break;

	case 'T':
		if (parse_action(arg, &status_temp_failure,
				 MAP(SMFIS_ACCEPT)|MAP(SMFIS_CONTINUE)|
				 MAP(SMFIS_DISCARD)|MAP(SMFIS_REJECT)|
				 MAP(SMFIS_TEMPFAIL)))
			argp_error(state,
				   "Bad value for '--bad-sender-action-action' option");
		break;
		
	case 'X':
		do_transcript = 1;
		break;
		
	case 'd':
		debug_level = atoi(arg);
		break;
		
	case 'D':
		smtp_domain = arg;
		break;

	case 't':
		test_mode = 1;
		log_to_stderr = 1;
		break;

	case OPTION_MILTER_TIMEOUT:
		if (smfi_settimeout(atoi(arg)) == MI_FAILURE) 
			argp_error(state, "Invalid timeout value"); 
		break;

	case OPTION_TIMEOUT:
		io_timeout = atoi(arg);
		break;
		
	case OPTION_RETRY:
		io_attempts = atoi(arg);
		break;
		
	case 's':
		log_to_stderr = 1;
		break;

	case OPTION_SHOW_DEFAULTS:
		printf("socket: %s\n", DEFAULT_SOCKET);
		printf("user: %s\n", DEFAULT_USER);
		printf("pidfile: %s\n", DEFAULT_PIDFILE);
		break;

	case OPTION_FOREGROUND:
		foreground = 1;
		break;

	case OPTION_PIDFILE:
		if (arg[0] != '/')
			argp_error(state,
				   "Invalid pidfile name: must be absolute");
		else
			pidfile = arg;
		break;
		
	case ARGP_KEY_FINI:
		if (!smtp_domain)
			mu_get_host_name(&smtp_domain);
		if (!log_to_stderr) {
			openlog(program_invocation_short_name,
				LOG_PID, log_facility);
			mu_error_set_print(syslog_error_printer);
		}
		break;

	case 'r':
		force_remove = 1;
		break;

	case 'u':
		user = arg;
		break;

	case OPTION_DOMAIN_FILE:
		read_domain_file(arg);
		break;
		
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static const char *capa[] = {
	"logging",
	NULL
};

static struct argp argp = {
	options,
	parse_opt,
	args_doc,
	doc,
	NULL,
	NULL,
	NULL
};


static void
check_local_portspec(char *p)
{
	struct stat st;
	
	if (stat(p, &st)) {
		if (errno == ENOENT)
			return; /* That's OK. The socket must not exist */
		mu_error("cannot stat file `%s': %s",
			 p, mu_strerror(errno));
		exit(EX_SOFTWARE);
	} else if (!force_remove) {
		mu_error("File `%s' already exist. Remove it or use --remove option.", p);
		exit(EX_SOFTWARE);
	}

	if (unlink (p)) {
		mu_error("cannot unlink file `%s': %s", p, mu_strerror(errno));
		exit(EX_SOFTWARE);
	}
}
		
int
switch_to_privs(uid_t uid, gid_t gid)
{
	int rc = 0;
	gid_t emptygidset[1];

	/* Reset group permissions */
	emptygidset[0] = gid ? gid : getegid();
	if (geteuid() == 0 && setgroups(1, emptygidset)) {
		mu_error("setgroups(1, %lu) failed: %s",
		         (unsigned long) emptygidset[0],
			 mu_strerror(errno));
		rc = 1;
	}

	/* Switch to the user's gid. On some OSes the effective gid must
	   be reset first */

#if defined(HAVE_SETEGID)
	if ((rc = setegid(gid)) < 0)
		mu_error("setegid(%lu) failed: %s",
			(unsigned long) gid, mu_strerror(errno));
#elif defined(HAVE_SETREGID)
	if ((rc = setregid(gid, gid)) < 0)
		mu_error("setregid(%lu,%lu) failed: %s",
		         (unsigned long) gid, (unsigned long) gid,
			 mu_strerror(errno));
#elif defined(HAVE_SETRESGID)
	if ((rc = setresgid(gid, gid, gid)) < 0)
		mu_error("setresgid(%lu,%lu,%lu) failed: %s",
		         (unsigned long) gid,
		         (unsigned long) gid,
		         (unsigned long) gid,
			 mu_strerror(errno));
#endif

	if (rc == 0 && gid != 0) {
		if ((rc = setgid(gid)) < 0 && getegid() != gid) 
			mu_error("setgid(%lu) failed: %s",
				 (unsigned long) gid, mu_strerror(errno));
		if (rc == 0 && getegid() != gid) {
			mu_error("cannot set effective gid to %lu",
			         (unsigned long) gid);
			rc = 1;
		}
	}

	/* Now reset uid */
	if (rc == 0 && uid != 0) {
		uid_t euid;

		if (setuid(uid)
		    || geteuid() != uid
		    || (getuid() != uid
			&& (geteuid() == 0 || getuid() == 0))) {
			
#if defined(HAVE_SETREUID)
			if (geteuid() != uid) {
				if (setreuid(uid, -1) < 0) { 
					mu_error("setreuid(%lu,-1) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
				if (setuid(uid) < 0) {
					mu_error("second setuid(%lu) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
			} else
#endif
				{
					mu_error("setuid(%lu) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
		}
	
		euid = geteuid();
		if (uid != 0 && setuid(0) == 0) {
			mu_error("seteuid(0) succeeded when it should not");
			rc = 1;
		} else if (uid != euid && setuid(euid) == 0) {
			mu_error("cannot drop non-root setuid privileges");
			rc = 1;
		}

	}

	return rc;
}

void
check_pidfile(char *name)
{
	unsigned long pid;
	FILE *fp = fopen(name, "r");
	if (!fp) {
		if (errno == ENOENT)
			return;
		mu_error("Cannot open pidfile `%s': %s",
			 name, mu_strerror(errno));
		exit(EX_TEMPFAIL);
	}
	if (fscanf(fp, "%lu", &pid) != 1) {
		mu_error("Cannot get pid from pidfile `%s'", name);
	} else {
		if (kill(pid, 0) == 0) {
			mu_error("%s appears to run with pid %lu. If it does not, remove `%s' and retry.",
				 program_invocation_short_name,
				 pid,
				 name);
			exit(EX_USAGE);
		}
	}
	fclose(fp);
	if (unlink(pidfile)) {
		mu_error("Cannot unlink pidfile `%s'",
			 name, mu_strerror(errno));
		exit(EX_USAGE);
	}
}

static char *license =
    "   This is mailfromd filter.\n"
    "   Copyright (C) 2005, Sergey Poznyakoff\n"
    "\n"
    "   ls is free software; you can redistribute it and/or modify\n"
    "   it under the terms of the GNU General Public License as published by\n"
    "   the Free Software Foundation; either version 2 of the License, or\n"
    "   (at your option) any later version.\n"
    "\n"
    "   GNU Mailutils is distributed in the hope that it will be useful,\n"
    "   but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "   GNU General Public License for more details.\n"
    "\n"
    "   You should have received a copy of the GNU General Public License along\n"
    "   with GNU Mailutils; if not, write to the Free Software Foundation,\n"
"   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA\n\n";

int
main(int argc, char **argv)
{
	int index;
	struct passwd *pw;

	mu_license_text = license;
	mu_argp_init (program_version, PACKAGE_BUGREPORT);
	mu_argp_parse (&argp, &argc, &argv, 0, capa, &index, NULL);

	if (test_mode) {
		int i;
		
		argv += index;
		argc -= index;

		if (argc % 2) {
			mu_error("Incorrect number of parameters");
			exit(EX_USAGE);
		}
		
		for (i = 0; i < argc; i += 2) {
			sfsistat stat = check_sender(NULL, argv[i], argv[i+1]);
			printf("%s @ %s => ", argv[i], argv[i+1]);
			print_stat(stat);
			printf("\n");
		}
		exit(EX_OK);
	}
				
	if (getuid() == 0) {
		if (!user) {
			mu_error("When running as root, --user option is mandatory.");
			exit(EX_USAGE);
		} else {
			pw = getpwnam(user);
			if (!pw) {
				mu_error("no such user: %s", user);
				error(EX_SOFTWARE);
			}
		}
	}

	if (pw && switch_to_privs(pw->pw_uid, pw->pw_gid))
		exit(EX_SOFTWARE);
	
	if (!foreground)
		check_pidfile(pidfile);
	
	if (strncasecmp(portspec, "unix:", 5) == 0) {
		check_local_portspec(portspec + 5);
		umask(007);
	} else if (strncasecmp(portspec, "local:", 6) == 0) {
		check_local_portspec(portspec + 6);
		umask(007);
	}

	if (smfi_setconn(portspec) == MI_FAILURE) {
		mu_error("smfi_setconn: %s", mu_strerror(errno));
		exit(EX_SOFTWARE);
	}
	
	if (smfi_register(smfilter) == MI_FAILURE) {
		mu_error("smfi_register failed");
		exit(EX_UNAVAILABLE);
	}

	if (!foreground) {
		daemon(0, 0);
		daemon_create_pidfile(pidfile);
	}

	return smfi_main();
}
